# -*- coding: utf-8 -*-
"""
Shim providing some notebook.nbextensions functions for versions < 4.2.0.

The main differences are:
 * check_nbextension public API function omitted as unused
 * all apps except BaseNBExtensionApp are omitted
 * some unused private API stuff omitted
 * all deprecated arguments are ignored, since we assume the caller knows what
   they're doing
 * docstrings limited to first line for brevity - we assume that if the caller
   is using the function, they're familiar with notebook 4.2 API
"""

# Original jupyter notebook source is
# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

from __future__ import print_function

import copy
import os
import shutil
import tarfile
import zipfile

from ipython_genutils.path import ensure_dir_exists
from ipython_genutils.py3compat import cast_unicode_py2
from ipython_genutils.tempdir import TemporaryDirectory
from jupyter_core.application import JupyterApp
from jupyter_core.paths import (
    ENV_CONFIG_PATH, ENV_JUPYTER_PATH, SYSTEM_CONFIG_PATH, SYSTEM_JUPYTER_PATH,
    jupyter_config_dir, jupyter_data_dir,
)
from notebook.nbextensions import (
    ArgumentConflict, __version__, _safe_is_tarfile,
)
from tornado.log import LogFormatter
from traitlets import Bool
from traitlets.config.manager import BaseJSONConfigManager
from traitlets.utils.importstring import import_item

try:
    from urllib.parse import urlparse  # Py3
    from urllib.request import urlretrieve
except ImportError:
    from urlparse import urlparse
    from urllib import urlretrieve

# Constants for pretty printing
# Window doesn't support coloring in the commandline
GREEN_OK = '\033[32mOK\033[0m' if os.name != 'nt' else 'ok'
RED_X = '\033[31m X\033[0m' if os.name != 'nt' else ' X'
GREEN_ENABLED = '\033[32m enabled \033[0m' if os.name != 'nt' else 'enabled '
RED_DISABLED = '\033[31mdisabled\033[0m' if os.name != 'nt' else 'disabled'

NBCONFIG_SECTIONS = ['common', 'notebook', 'tree', 'edit', 'terminal']

# -----------------------------------------------------------------------------
# Public API - large parts omitted
# -----------------------------------------------------------------------------


# check_nbextension omitted as unused


def install_nbextension(path, overwrite=False, symlink=False,
                        user=False, prefix=None, nbextensions_dir=None,
                        destination=None, logger=None, sys_prefix=False):
    """Install a Javascript extension for the notebook."""

    # the actual path to which we eventually installed
    full_dest = None

    nbext = _get_nbextension_dir(
        user=user, sys_prefix=sys_prefix, prefix=prefix,
        nbextensions_dir=nbextensions_dir)
    # make sure nbextensions dir exists
    ensure_dir_exists(nbext)

    # forcing symlink parameter to False if os.symlink does not exist (e.g.,
    # on Windows machines running python 2)
    if not hasattr(os, 'symlink'):
        symlink = False

    if isinstance(path, (list, tuple)):
        raise TypeError("path must be a string pointing to a single extension "
                        "to install; call this function multiple times to "
                        "install multiple extensions")

    path = cast_unicode_py2(path)

    if path.startswith(('https://', 'http://')):
        if symlink:
            raise ValueError("Cannot symlink from URLs")
        # Given a URL, download it
        with TemporaryDirectory() as td:
            filename = urlparse(path).path.split('/')[-1]
            local_path = os.path.join(td, filename)
            if logger:
                logger.info("Downloading: %s -> %s" % (path, local_path))
            urlretrieve(path, local_path)
            # now install from the local copy
            full_dest = install_nbextension(
                local_path, overwrite=overwrite, symlink=symlink,
                nbextensions_dir=nbext, destination=destination, logger=logger)
    elif path.endswith('.zip') or _safe_is_tarfile(path):
        if symlink:
            raise ValueError("Cannot symlink from archives")
        if destination:
            raise ValueError("Cannot give destination for archives")
        if logger:
            logger.info("Extracting: %s -> %s" % (path, nbext))

        if path.endswith('.zip'):
            archive = zipfile.ZipFile(path)
        elif _safe_is_tarfile(path):
            archive = tarfile.open(path)
        archive.extractall(nbext)
        archive.close()
        # TODO: what to do here
        full_dest = None
    else:
        if not destination:
            destination = os.path.basename(path)
        destination = cast_unicode_py2(destination)
        full_dest = os.path.normpath(os.path.join(nbext, destination))
        if overwrite and os.path.lexists(full_dest):
            if logger:
                logger.info("Removing: %s" % full_dest)
            if os.path.isdir(full_dest) and not os.path.islink(full_dest):
                shutil.rmtree(full_dest)
            else:
                os.remove(full_dest)

        if symlink:
            path = os.path.abspath(path)
            if not os.path.exists(full_dest):
                if logger:
                    logger.info("Symlinking: %s -> %s" % (full_dest, path))
                os.symlink(path, full_dest)
        elif os.path.isdir(path):
            # end in path separator
            path = os.path.join(os.path.abspath(path), '')
            for parent, dirs, files in os.walk(path):
                dest_dir = os.path.join(full_dest, parent[len(path):])
                if not os.path.exists(dest_dir):
                    if logger:
                        logger.info("Making directory: %s" % dest_dir)
                    os.makedirs(dest_dir)
                for file in files:
                    src = os.path.join(parent, file)
                    dest_file = os.path.join(dest_dir, file)
                    _maybe_copy(src, dest_file, logger=logger)
        else:
            src = path
            _maybe_copy(src, full_dest, logger=logger)

    return full_dest


def install_nbextension_python(module, overwrite=False, symlink=False,
                               user=False, sys_prefix=False, prefix=None,
                               nbextensions_dir=None, logger=None):
    """Install an nbextension bundled in a Python package."""
    m, nbexts = _get_nbextension_metadata(module)
    base_path = os.path.split(m.__file__)[0]

    full_dests = []

    for nbext in nbexts:
        src = os.path.join(base_path, nbext['src'])
        dest = nbext['dest']

        if logger:
            logger.info("Installing %s -> %s" % (src, dest))
        full_dest = install_nbextension(
            src, overwrite=overwrite, symlink=symlink,
            user=user, sys_prefix=sys_prefix, prefix=prefix,
            nbextensions_dir=nbextensions_dir,
            destination=dest, logger=logger
        )
        validate_nbextension_python(nbext, full_dest, logger)
        full_dests.append(full_dest)

    return full_dests


def uninstall_nbextension(dest, require=None, user=False, sys_prefix=False,
                          prefix=None, nbextensions_dir=None, logger=None):
    """Uninstall a Javascript extension of the notebook."""
    nbext = _get_nbextension_dir(user=user, sys_prefix=sys_prefix,
                                 prefix=prefix,
                                 nbextensions_dir=nbextensions_dir)
    dest = cast_unicode_py2(dest)
    full_dest = os.path.join(nbext, dest)
    if os.path.lexists(full_dest):
        if logger:
            logger.info("Removing: %s" % full_dest)
        if os.path.isdir(full_dest) and not os.path.islink(full_dest):
            shutil.rmtree(full_dest)
        else:
            os.remove(full_dest)

    # Look through all of the config sections making sure that the nbextension
    # doesn't exist.
    config_dir = os.path.join(
        _get_config_dir(user=user, sys_prefix=sys_prefix), 'nbconfig')
    cm = BaseJSONConfigManager(config_dir=config_dir)
    if require:
        for section in NBCONFIG_SECTIONS:
            cm.update(section, {"load_extensions": {require: None}})


def uninstall_nbextension_python(module, user=False, sys_prefix=False,
                                 prefix=None, nbextensions_dir=None,
                                 logger=None):
    """Uninstall an nbextension bundled in a Python package."""
    m, nbexts = _get_nbextension_metadata(module)
    for nbext in nbexts:
        dest = nbext['dest']
        require = nbext['require']
        if logger:
            logger.info("Uninstalling {} {}".format(dest, require))
        uninstall_nbextension(dest, require, user=user, sys_prefix=sys_prefix,
                              prefix=prefix, nbextensions_dir=nbextensions_dir,
                              logger=logger)


def _set_nbextension_state(section, require, state,
                           user=True, sys_prefix=False, logger=None):
    """Set whether the section's frontend should require the nbextension."""
    user = False if sys_prefix else user
    config_dir = os.path.join(
        _get_config_dir(user=user, sys_prefix=sys_prefix), 'nbconfig')
    cm = BaseJSONConfigManager(config_dir=config_dir)
    if logger:
        logger.info("{} {} extension {}...".format(
            "Enabling" if state else "Disabling",
            section,
            require
        ))
    cm.update(section, {"load_extensions": {require: state}})

    validate_nbextension(require, logger=logger)

    return cm.get(section).get(require) == state


def _set_nbextension_state_python(state, module, user, sys_prefix,
                                  logger=None):
    """Enable or disable some nbextensions stored in a Python package."""
    m, nbexts = _get_nbextension_metadata(module)
    return [_set_nbextension_state(section=nbext["section"],
                                   require=nbext["require"],
                                   state=state,
                                   user=user, sys_prefix=sys_prefix,
                                   logger=logger)
            for nbext in nbexts]


def enable_nbextension(section, require, user=True, sys_prefix=False,
                       logger=None):
    """Enable a named nbextension."""
    return _set_nbextension_state(section=section, require=require,
                                  state=True,
                                  user=user, sys_prefix=sys_prefix,
                                  logger=logger)


def disable_nbextension(section, require, user=True, sys_prefix=False,
                        logger=None):
    """Disable a named nbextension."""
    return _set_nbextension_state(section=section, require=require,
                                  state=False,
                                  user=user, sys_prefix=sys_prefix,
                                  logger=logger)


def enable_nbextension_python(module, user=True, sys_prefix=False,
                              logger=None):
    """Enable some nbextensions associated with a Python module."""
    return _set_nbextension_state_python(True, module, user, sys_prefix,
                                         logger=logger)


def disable_nbextension_python(module, user=True, sys_prefix=False,
                               logger=None):
    """Disable some nbextensions associated with a Python module."""
    return _set_nbextension_state_python(False, module, user, sys_prefix,
                                         logger=logger)


def validate_nbextension(require, logger=None):
    """Validate a named nbextension."""
    warnings = []
    infos = []

    js_exists = False
    for exts in _nbextension_dirs():
        # Does the Javascript entrypoint actually exist on disk?
        js = u"{}.js".format(os.path.join(exts, *require.split("/")))
        js_exists = os.path.exists(js)
        if js_exists:
            break

    require_tmpl = u"        - require? {} {}"
    if js_exists:
        infos.append(require_tmpl.format(GREEN_OK, require))
    else:
        warnings.append(require_tmpl.format(RED_X, require))

    if logger:
        if warnings:
            logger.warning(u"      - Validating: problems found:")
            for msg in warnings:
                logger.warning(msg)
            for msg in infos:
                logger.info(msg)
        else:
            logger.info(u"      - Validating: {}".format(GREEN_OK))

    return warnings


def validate_nbextension_python(spec, full_dest, logger=None):
    """Assess the health of an installed nbextension."""
    infos = []
    warnings = []

    section = spec.get("section", None)
    if section in NBCONFIG_SECTIONS:
        infos.append(u"  {} section: {}".format(GREEN_OK, section))
    else:
        warnings.append(u"  {}  section: {}".format(RED_X, section))

    require = spec.get("require", None)
    if require is not None:
        require_path = os.path.join(
            full_dest[0:-len(spec["dest"])],
            u"{}.js".format(require))
        if os.path.exists(require_path):
            infos.append(u"  {} require: {}".format(GREEN_OK, require_path))
        else:
            warnings.append(u"  {}  require: {}".format(RED_X, require_path))

    if logger:
        if warnings:
            logger.warning("- Validating: problems found:")
            for msg in warnings:
                logger.warning(msg)
            for msg in infos:
                logger.info(msg)
            logger.warning(u"Full spec: {}".format(spec))
        else:
            logger.info(u"- Validating: {}".format(GREEN_OK))

    return warnings


# -----------------------------------------------------------------------------
# Applications. Many omitted from notebook version.
# -----------------------------------------------------------------------------


class BaseNBExtensionApp(JupyterApp):
    """Base nbextension installer app"""
    _log_formatter_cls = LogFormatter
    version = __version__

    flags = copy.deepcopy(JupyterApp.flags)
    flags.update({
        'user': ({
            'BaseNBExtensionApp': {
                'user': True,
            }}, 'Apply the operation only for the given user'
        ),
        'system': ({
            'BaseNBExtensionApp': {
                'user': False,
                'sys_prefix': False,
            }}, 'Apply the operation system-wide'
        ),
        'sys-prefix': ({
            'BaseNBExtensionApp': {
                'sys_prefix': True,
            }}, ('Use sys.prefix as the prefix for configuration operations ' +
                 'and installing nbextensions (for environments, packaging)')
        ),
        'py': ({
            'BaseNBExtensionApp': {
                'python': True,
            }}, 'Install from a Python package'
        )
    })
    flags['python'] = flags['py']
    flags.pop('y', None)
    flags.pop('generate-config', None)

    user = Bool(False, config=True, help="Whether to do a user install")
    sys_prefix = Bool(False, config=True,
                      help="Use the sys.prefix as the prefix")
    python = Bool(False, config=True, help="Install from a Python package")

    # stuff about verbose from notebook version omitted

    def _log_format_default(self):
        """A default format for messages"""
        return '%(message)s'

# -----------------------------------------------------------------------------
# Private API
# -----------------------------------------------------------------------------


def _should_copy(src, dest, logger=None):
    """Return whether a file should be copied."""
    if not os.path.exists(dest):
        return True
    if os.stat(src).st_mtime - os.stat(dest).st_mtime > 1e-6:
        # we add a fudge factor to work around a bug in python 2.x
        # that was fixed in python 3.x: http://bugs.python.org/issue12904
        if logger:
            logger.warn("Out of date: %s" % dest)
        return True
    if logger:
        logger.info("Up to date: %s" % dest)
    return False


def _maybe_copy(src, dest, logger=None):
    """Copy a file if it needs updating."""
    if _should_copy(src, dest, logger=logger):
        if logger:
            logger.info("Copying: %s -> %s" % (src, dest))
        shutil.copy2(src, dest)


def _get_nbextension_dir(user=False, sys_prefix=False, prefix=None,
                         nbextensions_dir=None):
    """Return the nbextension directory specified."""
    if sum(map(bool, [user, prefix, nbextensions_dir, sys_prefix])) > 1:
        raise ArgumentConflict(
            "cannot specify more than one of user, sys_prefix, prefix, "
            "or nbextensions_dir")
    if user:
        nbext = os.path.join(jupyter_data_dir(), u'nbextensions')
    elif sys_prefix:
        nbext = os.path.join(ENV_JUPYTER_PATH[0], u'nbextensions')
    elif prefix:
        nbext = os.path.join(prefix, 'share', 'jupyter', 'nbextensions')
    elif nbextensions_dir:
        nbext = nbextensions_dir
    else:
        nbext = os.path.join(SYSTEM_JUPYTER_PATH[0], 'nbextensions')
    return nbext


def _nbextension_dirs():
    """The possible locations of nbextensions."""
    return [
        os.path.join(jupyter_data_dir(), u'nbextensions'),
        os.path.join(ENV_JUPYTER_PATH[0], u'nbextensions'),
        os.path.join(SYSTEM_JUPYTER_PATH[0], 'nbextensions')
    ]


def _get_config_dir(user=False, sys_prefix=False):
    """Get the location of config files for the current context."""
    user = False if sys_prefix else user
    if user and sys_prefix:
        raise ArgumentConflict(
            "Cannot specify more than one of user or sys_prefix")
    if user:
        nbext = jupyter_config_dir()
    elif sys_prefix:
        nbext = ENV_CONFIG_PATH[0]
    else:
        nbext = SYSTEM_CONFIG_PATH[0]
    return nbext


def _get_nbextension_metadata(module):
    """Get the list of nbextension paths associated with a Python module."""
    m = import_item(module)
    if not hasattr(m, '_jupyter_nbextension_paths'):
        raise KeyError(
            'The Python module {} is not a valid nbextension'.format(module))
    nbexts = m._jupyter_nbextension_paths()
    return m, nbexts



# -*- coding: utf-8 -*-
"""Shim providing notebook.nbextensions stuff from 4.2 for earlier versions."""

try:
    from notebook.nbextensions import (
        # constants
        GREEN_ENABLED, GREEN_OK, NBCONFIG_SECTIONS, RED_DISABLED, RED_X,
        # Apps & classes
        ArgumentConflict, BaseNBExtensionApp,
        # public API functions
        _set_nbextension_state, _set_nbextension_state_python,
        disable_nbextension, disable_nbextension_python,
        enable_nbextension, enable_nbextension_python,
        install_nbextension, install_nbextension_python,
        uninstall_nbextension, uninstall_nbextension_python,
        validate_nbextension, validate_nbextension_python,
        # private API functions
        _get_nbextension_dir, _nbextension_dirs, _get_config_dir,
        _get_nbextension_metadata,
    )
except ImportError:
    from ._compat.nbextensions import (
        # constants
        GREEN_ENABLED, GREEN_OK, NBCONFIG_SECTIONS, RED_DISABLED, RED_X,
        # Apps & classes
        ArgumentConflict, BaseNBExtensionApp,
        # public API functions
        _set_nbextension_state, _set_nbextension_state_python,
        disable_nbextension, disable_nbextension_python,
        enable_nbextension, enable_nbextension_python,
        install_nbextension, install_nbextension_python,
        uninstall_nbextension, uninstall_nbextension_python,
        validate_nbextension, validate_nbextension_python,
        # private API functions
        _get_nbextension_dir, _nbextension_dirs, _get_config_dir,
        _get_nbextension_metadata,
    )

__all__ = [
    # constants
    'GREEN_ENABLED', 'GREEN_OK', 'NBCONFIG_SECTIONS', 'RED_DISABLED', 'RED_X',
    # Apps & classes
    'ArgumentConflict', 'BaseNBExtensionApp',
    # public API functions
    '_set_nbextension_state', '_set_nbextension_state_python',
    'disable_nbextension', 'disable_nbextension_python',
    'enable_nbextension', 'enable_nbextension_python',
    'install_nbextension', 'install_nbextension_python',
    'uninstall_nbextension', 'uninstall_nbextension_python',
    'validate_nbextension', 'validate_nbextension_python',
    # private API functions
    '_get_nbextension_dir', '_nbextension_dirs', '_get_config_dir',
    '_get_nbextension_metadata',
]



# coding: utf-8
"""
Common application classes for jupyter_contrib.

Including the root `jupyter-contrib` command.
"""

from __future__ import print_function

import sys

import pkg_resources
from jupyter_core.application import JupyterApp

from jupyter_contrib_core import __version__


class JupyterContribApp(JupyterApp):
    """Root level jupyter_contrib app."""

    name = 'jupyter contrib'
    version = __version__
    description = (
        'community-contributed spice for Jupyter Interactive Computing')

    def __init__(self, *args, **kwargs):
        self._refresh_subcommands()
        super(JupyterContribApp, self).__init__(*args, **kwargs)

    def _refresh_subcommands(self):
        """
        Finds subcommands which have registered entry points.

        Each entry point is a function which returns a subcommands-style dict,
        where the keys are the name of the subcommand, and the values are
        2-tuples containing the sub-application class, and a description of the
        subcommand's action.
        """
        group = 'jupyter_contrib_core.app.subcommands'
        new_subcommands = {}
        # import ipdb; ipdb.set_trace()
        for entrypoint in pkg_resources.iter_entry_points(group=group):
            get_subcommands_dict = entrypoint.load()
            new_subcommands.update(get_subcommands_dict())
        self.subcommands.clear()
        self.subcommands.update(new_subcommands)

    def start(self):
        """Perform the App's actions as configured"""
        super(JupyterContribApp, self).start()

        # The above should have called a subcommand and raised NoStart; if we
        # get here, it didn't, so we should self.log.info a message.
        self.print_help()
        subcmds = ", ".join(sorted(self.subcommands))
        sys.exit("Please supply at least one subcommand: %s" % subcmds)

main = JupyterContribApp.launch_instance

if __name__ == '__main__':  # pragma: no cover
    main()



# -*- coding: utf-8 -*-
# - Copyright (c) 2016-, jupyter-contrib development team

"""Core functionality for jupyter-contrib projects."""

from __future__ import unicode_literals

__version__ = '0.3.0'



# -*- coding: utf-8 -*-
"""Functions for patching jupyter env vars & paths."""

from __future__ import (
    absolute_import, division, print_function, unicode_literals,
)

import os
import shutil
import sys
import tempfile

import jupyter_core.paths

from jupyter_contrib_core.notebook_compat import nbextensions
from jupyter_contrib_core.testing_utils import stringify_env

try:
    from unittest.mock import patch
except ImportError:
    from mock import patch  # py2


def make_dirs(test_dir, base_dir):
    """Return a dict of root, config and data directory paths."""
    dirs = {
        'root': os.path.join(test_dir, base_dir),
        'conf': os.path.join(test_dir, base_dir, 'config'),
        'data': os.path.join(test_dir, base_dir, 'data'),
    }
    if not os.path.exists(dirs['root']):
        os.makedirs(dirs['root'])
    return dirs


def patch_jupyter_dirs():
    """
    Patch jupyter paths to use temporary directories.

    This just creates the patches and directories, caller is still
    responsible for starting & stopping patches, and removing temp dir when
    appropriate.
    """
    test_dir = tempfile.mkdtemp(prefix='jupyter_')
    jupyter_dirs = {name: make_dirs(test_dir, name) for name in (
        'user_home', 'env_vars', 'system', 'sys_prefix', 'custom', 'server')}
    jupyter_dirs['root'] = test_dir

    for name in ('notebook', 'runtime'):
        d = jupyter_dirs['server'][name] = os.path.join(
            test_dir, 'server', name)
        if not os.path.exists(d):
            os.makedirs(d)

    # patch relevant environment variables
    jupyter_patches = []
    jupyter_patches.append(
        patch.dict('os.environ', stringify_env({
            'HOME': jupyter_dirs['user_home']['root'],
            'JUPYTER_CONFIG_DIR': jupyter_dirs['env_vars']['conf'],
            'JUPYTER_DATA_DIR': jupyter_dirs['env_vars']['data'],
            'JUPYTER_RUNTIME_DIR': jupyter_dirs['server']['runtime'],
        })))

    # patch jupyter path variables in various modules
    # find the appropriate modules to patch according to compat.
    # Should include either
    # notebook.nbextensions
    # or
    # jupyter_contrib_core.notebook_compat._compat.nbextensions
    modules_to_patch = (
        jupyter_core.paths,
        sys.modules[nbextensions._get_config_dir.__module__])
    path_patches = dict(
        SYSTEM_CONFIG_PATH=[jupyter_dirs['system']['conf']],
        ENV_CONFIG_PATH=[jupyter_dirs['sys_prefix']['conf']],
        SYSTEM_JUPYTER_PATH=[jupyter_dirs['system']['data']],
        ENV_JUPYTER_PATH=[jupyter_dirs['sys_prefix']['data']],
    )
    for mod in modules_to_patch:
        applicable_patches = {
            attrname: newval for attrname, newval in path_patches.items()
            if hasattr(mod, attrname)}
        jupyter_patches.append(
            patch.multiple(mod, **applicable_patches))

    def remove_jupyter_dirs():
        """Remove all temporary directories created."""
        shutil.rmtree(test_dir)

    return jupyter_patches, jupyter_dirs, remove_jupyter_dirs



# -*- coding: utf-8 -*-
"""Test utilities."""

from __future__ import (
    absolute_import, division, print_function, unicode_literals,
)

import logging
import os
import re
import sys
from threading import RLock

from traitlets.config.application import LevelFormatter
from traitlets.traitlets import default

try:
    from notebook.tests.test_notebookapp import raise_on_bad_version
except ImportError:
    pep440re = re.compile(
        r'^'
        r'([1-9]\d*!)?(0|[1-9]\d*)(.(0|[1-9]\d*))*'
        r'((a|b|rc)(0|[1-9]\d*))?'
        r'(\.post(0|[1-9]\d*))?'
        r'(\.dev(0|[1-9]\d*))?'
        r'$'
    )

    def raise_on_bad_version(version):
        if not pep440re.match(version):
            raise ValueError(
                "Versions String apparently does not match Pep 440 "
                "specification, which might lead to sdist and wheel being "
                "seen as 2 different release. "
                "E.g: do not use dots for beta/alpha/rc markers."
            )


def stringify_env(env):
    """
    Convert environment vars dict to str: str (not unicode) for py2 on Windows.

    Python 2 on Windows doesn't handle Unicode objects in environment, even if
    they can be converted to ASCII string, which can cause problems for
    subprocess calls in modules importing unicode_literals from future.
    """
    if sys.version_info[0] < 3 and os.name == 'nt':
        return {str(key): str(val) for key, val in env.iteritems()}
    return env


class GlobalMemoryHandler(logging.Handler):
    """
    A MemoryHandler which uses a single buffer across all instances.

    In addition, will only flush logs when explicitly called to.
    """

    _buffer = None  # used as a class-wide attribute
    _lock = None  # used as a class-wide attribute

    @classmethod
    def _setup_class(cls):
        if cls._lock is None:
            cls._lock = RLock()
        if cls._buffer is None:
            with cls._lock:
                cls._buffer = []

    def __init__(self, target):
        logging.Handler.__init__(self)
        self.target = target
        self._setup_class()

    def emit(self, record):
        """
        Emit a record.

        Append the record and its target to the buffer.
        Don't check shouldFlush like regular MemoryHandler does.
        """
        self.__class__._buffer.append((record, self.target))

    @classmethod
    def flush_to_target(cls):
        """
        Sending the buffered records to their respective targets.

        The class-wide record buffer is also cleared by this operation.
        """
        with cls._lock:
            for record, target in cls._buffer:
                target.handle(record)
            cls.clear_buffer()

    @classmethod
    def clear_buffer(cls):
        with cls._lock:
            cls._buffer = []

    @classmethod
    def rotate_buffer(cls, num_places=1):
        with cls._lock:
            cls._buffer = cls._buffer[-num_places:] + cls._buffer[:-num_places]

    def close(self):
        """Close the handler."""
        try:
            self.flush()
        finally:
            logging.Handler.close(self)


def wrap_logger_handlers(logger):
    """Wrap a logging handler in a GlobalMemoryHandler."""
    # clear original log handlers, saving a copy
    handlers_to_wrap = logger.handlers
    logger.handlers = []
    # wrap each one
    for handler in handlers_to_wrap:
        if isinstance(handler, GlobalMemoryHandler):
            wrapping_handler = handler
        else:
            wrapping_handler = GlobalMemoryHandler(target=handler)
        logger.addHandler(wrapping_handler)
    return logger


def get_logger(name=__name__, log_level=logging.DEBUG):
    """
    Return a logger with a default StreamHandler.

    Adapted from
        tratilets.config.application.Application._log_default
    """
    log = logging.getLogger(name)
    log.setLevel(log_level)
    log.propagate = False
    _log = log  # copied from Logger.hasHandlers() (new in Python 3.2)
    while _log:
        if _log.handlers:
            return log
        if not _log.propagate:
            break
        else:
            _log = _log.parent
    if sys.executable.endswith('pythonw.exe'):
        # this should really go to a file, but file-logging is only
        # hooked up in parallel applications
        _log_handler = logging.StreamHandler(open(os.devnull, 'w'))
    else:
        _log_handler = logging.StreamHandler()
    _log_formatter = LevelFormatter(
        fmt='[%(levelname)1.1s %(asctime)s.%(msecs).03d %(name)s] %(message)s',
        datefmt='%H:%M:%S')
    _log_handler.setFormatter(_log_formatter)
    log.addHandler(_log_handler)
    return log


def get_wrapped_logger(*args, **kwargs):
    """Return a logger with StreamHandler wrapped in a GlobalMemoryHandler."""
    return wrap_logger_handlers(get_logger(*args, **kwargs))


def patch_traitlets_app_logs(klass):
    """
    Patch an App's default log method for use in nose tests.

    This is for use on subclasses of tratilets.config.application.Application
    and essentially removes handlers from the default logger, then sets it to
    propagate so that nose can capture the logs.
    """
    @default('log')
    def new_default_log(self):
        logger = super(klass, self)._log_default()
        # clear log handlers and propagate to root for nose to capture
        logger.propagate = True
        logger.handlers = []
        return logger
    klass._log_default = new_default_log
